/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef MODULE_H
#define MODULE_H

#include "SOUNDfile.H"
#include "segmentData.H"
#include "segmentTable.H"
 
// types being allowed for input and output parameters of modules
typedef enum {
  MAAATE_TYPE_SOUNDFILE = 0,
  MAAATE_TYPE_SEGMENTTABLE,
  MAAATE_TYPE_SEGMENTDATA,
  MAAATE_TYPE_BOOL,
  MAAATE_TYPE_INT,
  MAAATE_TYPE_REAL,
  MAAATE_TYPE_STRING,
} MaaateType;


typedef enum {
  MAAATE_CONSTRAINT_NONE = 0,
  MAAATE_CONSTRAINT_VALUE,
  MAAATE_CONSTRAINT_RANGE
} MaaateConstraintType;

#ifdef __cplusplus

#include <string>
#include <list>

using namespace std;

class PluginLibrary;
class Plugins;

//--------------------------------------------------------------------------


// class to contain all different possible parameter types
class ModuleParam {
public:
  // constructors
  ModuleParam (SOUNDfile * f) : sf(f), type(MAAATE_TYPE_SOUNDFILE) {};
  ModuleParam (SegmentTable * s) : st(s), type(MAAATE_TYPE_SEGMENTTABLE) {};
  ModuleParam (SegmentData * s) : sd(s), type(MAAATE_TYPE_SEGMENTDATA) {};
  ModuleParam (bool bb) : b(bb), type(MAAATE_TYPE_BOOL) {};
  ModuleParam (int ii) : i(ii), type(MAAATE_TYPE_INT) {};
  ModuleParam (double rr) : r(rr), type(MAAATE_TYPE_REAL) {};
  ModuleParam (char * ss) : s(ss), type(MAAATE_TYPE_STRING) {};

  // copy constructor
  ModuleParam (const ModuleParam& mp);

  // overload <<
  friend ostream& operator<< (ostream& os, const ModuleParam& mp);
  ModuleParam&    operator= (const ModuleParam & mp);

  friend bool   operator== (const ModuleParam& mp1, const ModuleParam& mp2);
  friend double operator- (ModuleParam const & mp1, ModuleParam const & mp2);

  friend bool operator< (const ModuleParam& mp1, const ModuleParam& mp2);
  friend bool operator> (const ModuleParam& mp1, const ModuleParam& mp2);
  bool isZero();

  MaaateType getType() const { return type; }

  bool set (SOUNDfile * spf);
  bool set (SegmentTable * sgt);
  bool set (SegmentData * sgd);
  bool set (bool bl);
  bool set (int it);
  bool set (double rl);
  bool set (char * sg);
  
  SOUNDfile    * get_sf() const { return sf; }
  SegmentTable * get_st() const { return st; }
  SegmentData  * get_sd() const { return sd; }
  bool           get_b()  const { return b; }
  int            get_i()  const { return i; }
  double         get_r()  const { return r; }
  char *         get_s()  const { return s; }
  
private:
  // parameter data
  SOUNDfile    * sf;
  SegmentTable * st;
  SegmentData  * sd;
  bool            b;
  int             i;
  double          r;
  char *          s;

  // which type is actually stored?
  MaaateType type;
};

typedef list<ModuleParam> ModuleParamList;


class ModuleParamRange {
public:
  ModuleParamRange (ModuleParam * lowerBound,
		    ModuleParam * upperBound,
		    ModuleParam * step):
    lB(lowerBound),uB(upperBound),s(step) {};

  // overload <<
  friend ostream& operator<< (ostream& os, const ModuleParamRange& mpr);

  ModuleParam * lowerBound() const { return lB; }
  ModuleParam * upperBound() const { return uB; }
  ModuleParam * step()       const { return s; }

  void setLowerBound (ModuleParam * low) { lB = low; }
  void setUpperBound (ModuleParam * up)  { uB = up; }
  void setStep       (ModuleParam * st)  { s  = st; }

private:
  ModuleParam * lB;
  ModuleParam * uB;
  ModuleParam * s;
};


union ModuleParamConstraintValue {
public:
  // constructors
  ModuleParamConstraintValue (ModuleParam * val) : value(val) {};
  ModuleParamConstraintValue (ModuleParamRange * ran) : range(ran) {};

  ModuleParam *      constraintValue () const { return value; }
  ModuleParamRange * constraintRange () const { return range; }

private:
  // either one of these is stored
  ModuleParam * value;
  ModuleParamRange * range;
};

class ModuleParamConstraint {
public:
  ModuleParamConstraint (MaaateConstraintType t,
			 ModuleParamConstraintValue * v) :
    type(t), constraint(v) {}
  MaaateConstraintType Type()   { return type; }
  ModuleParamConstraintValue * Value() { return constraint; }
  bool          within   (ModuleParam value);
  ModuleParam * closest  (ModuleParam param);
  double        distance (ModuleParam param);

  // overload <<
  friend ostream& operator<< (ostream& os, const ModuleParamConstraint& mpc);

private:
  MaaateConstraintType type;
  ModuleParamConstraintValue * constraint;
};

// specification and storage of constraints
class MaaateConstraint {
public:
  // initialise with empty list
  MaaateConstraint() { constraints.clear(); }

  // add a value to the list constraint
  void addConstraint (ModuleParam * value);

  // add a range constraint. Specify NULL values for bounds if they are
  // unbounded ;-P, specify NULL or zero for step if unused
  void addConstraintRange (ModuleParam * lowerBound,
			   ModuleParam * upperBound = NULL,
			   ModuleParam * step = NULL);
  void addConstraintRange (int lowerBound,
			   int upperBound,
			   int step = 1);
  void addConstraintRange (double lowerBound,
			   double upperBound,
			   double step = 0.0);
  void addConstraintGreaterThan(int i);
  void addConstraintGreaterThan(double d);
  void addConstraintLessThan(int i);
  void addConstraintLessThan(double d);

  void clear() {
    constraints.clear(); 
  }

  list<ModuleParamConstraint> * getConstraints() { return &constraints; }

  bool withinConstraints (ModuleParam value);
  ModuleParam * closestConstraintValue (ModuleParam value);

  // overload <<
  friend ostream& operator<< (ostream& os, MaaateConstraint& mc);

private:
  list<ModuleParamConstraint> constraints;
};

typedef  list<ModuleParamConstraint> ModuleParamConstraintList;

// class to contain a parameter specification
// no semantics on parameters yet included
class ModuleParamSpec {
public:
  ModuleParamSpec(string n,
		  string d,
		  MaaateType t,
		  ModuleParam * p,
		  MaaateConstraint * c = NULL) :
    mname(n), mdesc(d), mtype(t), mdefaultValue(p), mconstraint(c) {
    // XXX: delete whitespace from name!!!
  };

  string             name() const { return mname; }
  string             desc() const { return mdesc; }
  MaaateType         type() const { return mtype; }
  ModuleParam *      defaultValue() const { return mdefaultValue; }
  MaaateConstraint * constraint()   const { return mconstraint; }

  void setDefaultValue (ModuleParam * def) { mdefaultValue = def; }
  void setConstraint   (MaaateConstraint * con) { mconstraint = con; }

private:  
  string mname;  // for identification as a named argument; no whitespace!
  string mdesc;
  MaaateType mtype;

  // default value; might be overrun by suggested valued based on given params
  ModuleParam * mdefaultValue;
  MaaateConstraint * mconstraint;
};

typedef list<ModuleParamSpec> ModuleParamSpecList;

class Module;

typedef void (*ModuleInitFunc) (Module *);
typedef void (*ModuleDestroyFunc) (Module *);
typedef void (*ModuleResetFunc) (Module *);
typedef void (*ModuleSuggestValues) (Module * m,
				     list<ModuleParam> * paramsIn);
typedef list<ModuleParam>* (*ModuleApplyFunc) (Module * m,
					       list<ModuleParam> * paramsIn);
typedef list<ModuleParam>* (*ModuleDefaultFunc) (Module * m);

// default functions in the sane case
list<ModuleParam>* defaultD (Module * m);

// functions that do nothing for the insane case
list<ModuleParam>* defaultNoop (Module * m);
void suggestNoop (Module * m, list<ModuleParam> * paramsIn);
list<ModuleParam>* applyNoop (Module * m, list<ModuleParam> * params_in);
void resetNoop (Module * m);

// class to contain a module specification
// this is really just abstract and needs to be defined properly by a module
class Module {
  friend class PluginLibrary;
  friend class Plugins;

public:
  // constructor (performs initialisation with the initF, too)
  Module(ModuleInitFunc      initF,
	 ModuleDefaultFunc   defaultF,
	 ModuleSuggestValues suggestF,
	 ModuleApplyFunc     applyF,
	 ModuleResetFunc     resetF,
	 ModuleDestroyFunc   destroyF);

  // destructor
  ~Module(){
    if(destroyM != NULL) {
      destroyM(this);
    }
  }

  // member access
  bool   sane()       const { return saneM; }
  string name()       const { return modName; }
  string desc()       const { return modDesc; }
  string author()     const { return modAuthor; }
  string copyright()  const { return modCopyright; }
  string url()        const { return modUrl; }
  void set_name     (string nm) { modName = nm; }
  void set_desc     (string ds) { modDesc = ds; }
  void set_author   (string au) { modAuthor = au; }
  void set_copyright(string cp) { modCopyright = cp; }
  void set_url      (string ur) { modUrl = ur; }
  // specifications of input and output parameters
  list<ModuleParamSpec> * inputSpecs ()  { return &modParamInSpecs; }
  list<ModuleParamSpec> * outputSpecs () { return &modParamOutSpecs; }

  // function to provide default parameter values for a parameter list
  // which might be calculated from some of the parameters provided or
  // might be a single default value as stored with the parameter spec
  list<ModuleParam> * defaultValues () {
    return defaultM (this);
  };

  // function to suggest parameter values given that some of the input
  // parameters have already been set
  void suggestValues (list<ModuleParam> * paramsIn) {
    suggestM (this, paramsIn);
    checkConstraints (paramsIn);
  };
  
  // function to work on input parameters and return output parameters
  list<ModuleParam> * apply (list<ModuleParam> * paramsIn) {
    checkConstraints (paramsIn);
    return applyM (this, paramsIn);
  };

  // function to reset a module
  void reset () {
    resetM (this);
  };

  // function to ensure parameters are within the given constraints
  bool checkConstraints (list<ModuleParam> * paramsIn);

private:
  string modName;  // for identification in programs; no whitespace!
  string modDesc;
  string modAuthor;
  string modCopyright;
  string modUrl;
  list<ModuleParamSpec> modParamInSpecs;
  list<ModuleParamSpec> modParamOutSpecs;
  
  // function pointers
  ModuleDefaultFunc   defaultM;
  ModuleSuggestValues suggestM;
  ModuleApplyFunc     applyM;
  ModuleResetFunc     resetM;
  ModuleDestroyFunc   destroyM;

  // sanity checking
  bool saneM;

  // just used for internal purposes; shouldn't get changed by a plugin
  PluginLibrary * plib;
};

typedef list<Module> ModuleList;



extern "C" {
#else /* __cplusplus */
  typedef struct Module              Module;
  typedef struct ModuleParam         ModuleParam;
  typedef struct ModuleParamList     ModuleParamList;
  typedef struct ModuleParamConstraint  ModuleParamConstraint;
  typedef struct ModuleParamConstrainList ModuleParamConstraintList;
  typedef struct MaaateConstraint    MaaateConstraint;
  typedef struct ModuleParamSpec     ModuleParamSpec;
  typedef struct ModuleParamSpecList ModuleParamSpecList;
  typedef struct ModuleList          ModuleList;
  
  typedef void (*ModuleInitFunc)    (Module *);
  typedef void (*ModuleDestroyFunc) (Module *);
  typedef void (*ModuleResetFunc)   (Module *);
  typedef void (*ModuleSuggestValues) (Module * m,
				       ModuleParamList * paramsIn);
  typedef ModuleParamList * (*ModuleApplyFunc) (Module * m,
						ModuleParamList * paramsIn);
  typedef ModuleParamList * (*ModuleDefaultFunc) (Module * m);
  
#endif

  Module * maaateA_new_module(ModuleInitFunc      initF,
			      ModuleDefaultFunc   defaultF,
			      ModuleSuggestValues suggestF,
			      ModuleApplyFunc     applyF,
			      ModuleResetFunc     resetF,
			      ModuleDestroyFunc   destroyF);

  void maaateA_destroy_module(Module * m);

  int          maaateA_sane_module   (Module * m);
  const char * maaateA_name_module   (Module * m);
  const char * maaateA_desc_module   (Module * m);
  const char * maaateA_author_module (Module * m);
  const char * maaateA_copyright_module(Module * m);
  const char * maaateA_url_module    (Module * m);

  ModuleParamSpecList * maaateA_inputSpecs (Module * m);
  ModuleParamSpecList * maaateA_outputSpecs(Module * m);

  ModuleParamList * maaateA_default_values (Module * m);

  void maaateA_suggest_values (Module * m,
			       ModuleParamList * paramsIn);

  ModuleParamList * maaateA_apply (Module * m,
				   ModuleParamList * paramsIn);

  void maaateA_reset (Module * m);

  int  maaateA_check_constraints (Module * m,
				  ModuleParamList * paramsIn);

 /* wrapper functions for ModuleParamConstraint */

  /* wrapper functions for MaaateConstraint */
  ModuleParamConstraintList * maaateA_MaaateConstraint_getConstraints(MaaateConstraint * mc);
  unsigned int maaateA_MaaateConstraint_withinConstraints (MaaateConstraint * mc, ModuleParam * value);
  ModuleParam * maaateA_MaaateConstraint_closestConstraintValue (MaaateConstraint * mc, ModuleParam * value);

  /* wrapper functions for ModuleParamSpec */
  const char * maaateA_ModuleParamSpec_name (ModuleParamSpec * mps);
  const char * maaateA_ModuleParamSpec_desc (ModuleParamSpec * mps);
  MaaateType maaateA_ModuleParamSpec_type (ModuleParamSpec * mps);
  ModuleParam * maaateA_ModuleParamSpec_defaultvalue (ModuleParamSpec * mps);
  MaaateConstraint * maaateA_ModuleParamSpec_constraint (ModuleParamSpec * mps);


  /* wrapper functions for ModuleParam */
  MaaateType maaateA_ModuleParam_getType(ModuleParam * mp);

  void maaateA_ModuleParam_setSOUNDfile (ModuleParam * mp,
					SOUNDfile * sf);
  void maaateA_ModuleParam_setSegmentTable (ModuleParam * mp, 
					    SegmentTable * st);
  void maaateA_ModuleParam_setSegmentData (ModuleParam * mp,
					   SegmentData * sd);
  void maaateA_ModuleParam_setBool (ModuleParam * mp, int b);
  void maaateA_ModuleParam_setInt  (ModuleParam * mp, int i);
  void maaateA_ModuleParam_setReal (ModuleParam * mp, double r);
  void maaateA_ModuleParam_setString (ModuleParam * mp, char * s);

  SOUNDfile *    maaateA_ModuleParam_getSOUNDfile (ModuleParam * mp);
  SegmentTable * maaateA_ModuleParam_getSegmentTable (ModuleParam * mp);
  SegmentData *  maaateA_ModuleParam_getSegmentData (ModuleParam * mp);
  int            maaateA_ModuleParam_getBool (ModuleParam * mp);
  int            maaateA_ModuleParam_getInt  (ModuleParam * mp);
  double         maaateA_ModuleParam_getReal (ModuleParam * mp);
  char *         maaateA_ModuleParam_getString (ModuleParam * mp);


/* different list handling */
  ModuleParamConstraint * maaateA_ModuleParamConstraintList_nth (ModuleParamConstraintList * plist,
								 unsigned int n);

  unsigned int maaateA_ModuleParamConstraintList_length (ModuleParamConstraintList * plist);

  ModuleParamSpec * maaateA_ModuleParamSpecList_nth (ModuleParamSpecList * plist,
						 unsigned int n);

  unsigned int maaateA_ModuleParamSpecList_length (ModuleParamSpecList * plist);

  ModuleParam * maaateA_ModuleParamList_nth (ModuleParamList * plist,
					     unsigned int n);

  unsigned int maaateA_ModuleParamList_length (ModuleParamList * plist);

  Module * maaateA_ModuleList_nth (ModuleList * mlist,
				   unsigned int n);

  unsigned int maaateA_ModuleList_length (ModuleList * mlist);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
