/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef SEGMENTTABLE_H
#define SEGMENTTABLE_H

typedef enum { EXACT, BEFORE, AFTER } QuerySegmentType;

#ifdef __cplusplus

#include <vector>
#include <cmath>

#include "segmentData.H"

using namespace std;
 
//--------------------------------------------------------------------------
 
// class definition
class SegmentTable {
public:
  // different constructors
  SegmentTable() {segments.reserve(10);}

  // copy constructor
  SegmentTable (const SegmentTable& segment);

  // inversion of table segments (with loosing data)
  SegmentTable* invertTable (double endTime=0.0);

  // desctructor
  ~SegmentTable() {}

  // comparison between two SegmentTable objects with respect to their size
  friend bool operator< (const SegmentTable& st1, const SegmentTable& st2) {
    return (st1.segments < st2.segments);
  }

  // comparison between two SegmentTable objects with respect to their size
  friend bool operator> (const SegmentTable& st1, const SegmentTable& st2) {
    return (st1.segments > st2.segments);
  }

  // print all entries
  void print();
  void printPlain(bool negative);

  // overload <<
  friend ostream& operator<<(ostream& os, const SegmentTable& st);

  // overload =
  SegmentTable& operator= (const SegmentTable& st);
    
  // index operator
  SegmentData& operator[] (int i) {
    if ((i>=0) && (i<segments.size()))
      return segments[i];
    cerr << "MaaateA: SegmentTable: index out of range" << endl;
    return segments[0];
  }
    
  // returns the SegmentData that is at the specified time
  SegmentData * getSegment (double time, QuerySegmentType which);

  // returns the index of the (first) SegmentData that contains the time
  int getSegmentIndex (double time);
    
  // inserts/appends a SegmentData in the table and return its position
  vector<SegmentData>::iterator insert (const SegmentData& sd);
  void append (const SegmentData& sd);

  // inserts a SegmentTable in the table
  void insert (const SegmentTable& st);

  // deletes SegmentData at position ndx
  bool erase (int ndx);

  // deletes all entries
  void clear ();

  // positioning iterators
  vector<SegmentData>::iterator begin() { return segments.begin(); }
  vector<SegmentData>::iterator end() { return segments.end(); }

  int size() { return segments.size(); }
  bool empty() { return segments.empty(); }

  // calculate statistics over all segments
  double avgDuration();
  double minDuration();
  double maxDuration();

private:
  // audio block table
  vector<SegmentData> segments;
};

extern "C" {
#else /* __cplusplus */
  typedef struct SegmentTable SegmentTable;
#endif /* __cplusplus */

  SegmentTable * maaateA_st_new  ();
  SegmentTable * maaateA_st_copy (SegmentTable *st);
  SegmentTable * maaateA_st_invert_table(SegmentTable *st, double endTime);
  void           maaateA_st_delete (SegmentTable *st);

  int    maaateA_st_smaller (SegmentTable *st1, SegmentTable *st2);
  int    maaateA_st_bigger  (SegmentTable *st1, SegmentTable *st2);
  void   maaateA_st_print(SegmentTable * st);
  void   maaateA_st_print_plain(SegmentTable * st, int negative);
  SegmentData * maaateA_st_element_at(SegmentTable * st, int i);
  SegmentData * maaateA_st_get_segment(SegmentTable * st, 
				       double time, 
				       QuerySegmentType which);
  int           maaateA_st_get_segment_index(SegmentTable * st, double time);
  
  void   maaateA_st_append(SegmentTable * st, SegmentData * sd);
  void   maaateA_st_insert_st(SegmentTable * st, SegmentTable * st1);
  void   maaateA_st_erase(SegmentTable * st, int ndx);
  void   maaateA_st_clear(SegmentTable *st);
  int    maaateA_st_get_size(SegmentTable * st);
  int    maaateA_st_empty(SegmentTable *st);
  
  double maaateA_avg_duration(SegmentTable * st);
  double maaateA_min_duration(SegmentTable * st);
  double maaateA_max_duration(SegmentTable * st);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
